#!/usr/bin/env python3
'''Main DFT functions based on the DFT++ formulation.'''
import numpy as np
from numpy.random import Generator, SFC64
from scipy.linalg import eig, eigh, eigvalsh, inv, norm, sqrtm

from .gth import calc_Vnonloc
from .utils import diagprod, handle_spin_gracefully, pseudo_uniform
from .xc import get_vxc


def solve_poisson(atoms, n):
    '''Solve the Poisson equation.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        n (ndarray): Real-space electronic density.

    Returns:
        ndarray: Hartree field.
    '''
    # phi = -4 pi Linv(O(J(n)))
    return -4 * np.pi * atoms.Linv(atoms.O(atoms.J(n)))


def get_n_total(atoms, Y):
    '''Calculate the total electronic density.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        ndarray: Electronic density.
    '''
    # n = (IW) F (IW)dag
    Yrs = atoms.I(Y)
    n = np.zeros(len(atoms.r))
    for spin in range(atoms.Nspin):
        n += np.sum(atoms.f[spin] * np.real(Yrs[spin].conj() * Yrs[spin]), axis=1)
    return n


def get_n_spin(atoms, Y, n=None):
    '''Calculate the electronic density per spin channel.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Keyword Args:
        n (ndarray): Real-space electronic density.

    Returns:
        ndarray: Electronic density per spin.
    '''
    # Return the total density in the spin-paired case
    if n is not None and atoms.Nspin == 1:
        return np.atleast_2d(n)

    Yrs = atoms.I(Y)
    n = np.empty((atoms.Nspin, len(atoms.r)))
    for spin in range(atoms.Nspin):
        n[spin] = np.sum(atoms.f[spin] * np.real(Yrs[spin].conj() * Yrs[spin]), axis=1)
    return n


def get_n_single(atoms, Y):
    '''Calculate the single-electron densities.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        ndarray: Single-electron densities.
    '''
    Yrs = atoms.I(Y)
    n = np.empty((atoms.Nspin, len(atoms.r), atoms.Nstate))
    for spin in range(atoms.Nspin):
        n[spin] = atoms.f[spin] * np.real(Yrs[spin].conj() * Yrs[spin])
    return n


@handle_spin_gracefully
def orth(atoms, W):
    '''Orthogonalize coefficient matrix W.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: Orthogonalized wave functions.
    '''
    # Y = W (Wdag O(W))^-0.5
    return W @ inv(sqrtm(W.conj().T @ atoms.O(W)))


def get_grad(scf, spin, W, Y=None, n=None, phi=None, vxc=None):
    '''Calculate the energy gradient with respect to W.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        spin (int): Spin variable to track weather to calculate the gradient for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.
        n (ndarray): Real-space electronic density.
        phi (ndarray): Hartree field.
        vxc (ndarray): Exchange-correlation potential.

    Returns:
        ndarray: Gradient.
    '''
    atoms = scf.atoms
    F = np.diag(atoms.f[spin])
    HW = H(scf, spin, W, Y, n, phi, vxc)
    WHW = W[spin].conj().T @ HW
    # U = Wdag O(W)
    U = W[spin].conj().T @ atoms.O(W[spin])
    invU = inv(U)
    U12 = sqrtm(invU)
    # Htilde = U^-0.5 Wdag H(W) U^-0.5
    Ht = U12 @ WHW @ U12
    # grad E = H(W) - O(W) U^-1 (Wdag H(W)) (U^-0.5 F U^-0.5) + O(W) (U^-0.5 Q(Htilde F - F Htilde))
    return (HW - (atoms.O(W[spin]) @ invU) @ WHW) @ (U12 @ F @ U12) + \
        atoms.O(W[spin]) @ (U12 @ Q(Ht @ F - F @ Ht, U))


def H(scf, spin, W, Y=None, n=None, phi=None, vxc=None):
    '''Left-hand side of the eigenvalue equation.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        spin (int): Spin variable to track weather to calculate the gradient for spin up or down.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.
        n (ndarray): Real-space electronic density.
        phi (ndarray): Hartree field.
        vxc (ndarray): Exchange-correlation potential.

    Returns:
        ndarray: Hamiltonian applied on W.
    '''
    atoms = scf.atoms
    # One can calculate everything from W,
    # but one can also use already computed results to save time
    if Y is None:
        Y = orth(atoms, W)
    if n is None:
        n = get_n_total(atoms, Y)
    if phi is None:
        phi = solve_poisson(atoms, n)
    if vxc is None:
        n_spin = get_n_spin(atoms, Y, n)
        vxc = get_vxc(scf.xc, n_spin, atoms.Nspin)

    # We get the full potential in the functional definition (different to the DFT++ notation)
    # Normally Vxc = Jdag(O(J(exc))) + diag(exc') Jdag(O(J(n)))
    Vxc = atoms.Jdag(atoms.O(atoms.J(vxc[spin])))
    # Vkin = -0.5 L(W)
    Vkin_psi = -0.5 * atoms.L(W[spin])
    # Veff = Jdag(Vion) + Jdag(O(J(vxc))) + Jdag(O(phi))
    Veff = scf.Vloc + Vxc + atoms.Jdag(atoms.O(phi))
    Vnonloc_psi = calc_Vnonloc(scf, W[spin])
    # H = Vkin + Idag(diag(Veff))I + Vnonloc
    return Vkin_psi + atoms.Idag(diagprod(Veff, atoms.I(W[spin]))) + Vnonloc_psi


def Q(inp, U):
    '''Operator needed to calculate gradients with non-constant occupations.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        inp (ndarray): Coefficients input array.
        U (ndarray): Overlap of wave functions.

    Returns:
        ndarray: Q operator result.
    '''
    mu, V = eig(U)
    mu = mu[:, None]
    denom = np.sqrt(mu) @ np.ones((1, len(mu)))
    denom2 = denom + denom.conj().T
    return V @ ((V.conj().T @ inp @ V) / denom2) @ V.conj().T


def get_psi(scf, W, n=None):
    '''Calculate eigenstates from H.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        n (ndarray): Real-space electronic density.

    Returns:
        ndarray: Eigenstates in reciprocal space.
    '''
    atoms = scf.atoms
    Y = orth(atoms, W)
    psi = np.empty_like(Y)
    for spin in range(atoms.Nspin):
        mu = Y[spin].conj().T @ H(scf, spin, W=Y, n=n)
        _, D = eigh(mu)
        psi[spin] = Y[spin] @ D
    return psi


def get_epsilon(scf, W, n=None):
    '''Calculate eigenvalues from H.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        n (ndarray): Real-space electronic density.

    Returns:
        ndarray: Eigenvalues.
    '''
    atoms = scf.atoms
    Y = orth(atoms, W)
    epsilon = np.empty((atoms.Nspin, atoms.Nstate))
    for spin in range(atoms.Nspin):
        mu = Y[spin].conj().T @ H(scf, spin, W=Y, n=n)
        epsilon[spin] = np.sort(eigvalsh(mu))
    return epsilon


def guess_random(scf, complex=True):
    '''Generate random initial-guess coefficients as starting values.

    Args:
        scf: SCF object.

    Keyword Args:
        complex (bool): Use complex numbers for the random guess.

    Returns:
        ndarray: Initial-guess orthogonal wave functions in reciprocal space.
    '''
    atoms = scf.atoms

    seed = 42
    rng = Generator(SFC64(seed))
    if complex:
        W = rng.standard_normal((atoms.Nspin, len(atoms.G2c), atoms.Nstate)) + \
            1j * rng.standard_normal((atoms.Nspin, len(atoms.G2c), atoms.Nstate))
    else:
        W = rng.standard_normal((atoms.Nspin, len(atoms.G2c), atoms.Nstate))
    return orth(atoms, W)


def guess_gaussian(scf, complex=True):
    '''Generate initial-guess coefficients using normalized Gaussians as starting values.

    Args:
        scf: SCF object.

    Keyword Args:
        complex (bool): Use complex numbers for the random guess.

    Returns:
        ndarray: Initial-guess orthogonal wave functions in reciprocal space.
    '''
    atoms = scf.atoms
    # Start with randomized wave functions
    W = guess_random(scf, complex=complex)

    sigma = 0.5
    normal = (2 * np.pi * sigma**2)**(3 / 2)
    # Calculate a density from normalized Gauss functions
    n = np.zeros(len(atoms.r))
    for ia in range(atoms.Natoms):
        r = norm(atoms.r - atoms.X[ia], axis=1)
        n += atoms.Z[ia] * np.exp(-r**2 / (2 * sigma**2)) / normal
    # Calculate the eigenfunctions
    return get_psi(scf, W, n)


def guess_pseudo(scf, seed=1234):
    '''Generate initial-guess coefficients using pseudo-random starting values.

    Args:
        scf: SCF object.

    Keyword Args:
        seed (int): Seed to initialize the random number generator.

    Returns:
        ndarray: Initial-guess orthogonal wave functions in reciprocal space.
    '''
    atoms = scf.atoms
    W = pseudo_uniform((atoms.Nspin, len(atoms.G2c), atoms.Nstate), seed=seed)
    return orth(atoms, W)



#!/usr/bin/env python3
'''Modified Perdew-Wang LDA correlation.

Reference: Phys. Rev. B 45, 13244.
'''
from .lda_c_pw import lda_c_pw, lda_c_pw_spin


def lda_c_pw_mod(n, **kwargs):
    '''Modified Perdew-Wang parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW_MOD and ID 13 in Libxc.
    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    '''
    return lda_c_pw(n, a=0.0310907, **kwargs)


def lda_c_pw_mod_spin(n, zeta, **kwargs):
    '''Modified Perdew-Wang parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW_MOD and ID 13 in Libxc.
    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    '''
    return lda_c_pw_spin(n, zeta, a=(0.0310907, 0.01554535, 0.0168869),
                         fz0=1.709920934161365617563962776245, **kwargs)



#!/usr/bin/env python3
'''Vosko-Wilk-Nusair LDA correlation.

Reference: Phys. Rev. B 22, 3812.
'''
import numpy as np


def lda_c_vwn(n, exc_only=False, **kwargs):
    '''Vosko-Wilk-Nusair parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_VWN and ID 7 in Libxc.
    Reference: Phys. Rev. B 22, 3812.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: VWN correlation energy density and potential.
    '''
    a = 0.0310907
    b = 3.72744
    c = 12.9352
    x0 = -0.10498

    pi34 = (3 / (4 * np.pi))**(1 / 3)
    rs = pi34 / n**(1 / 3)

    q = np.sqrt(4 * c - b * b)
    f1 = 2 * b / q
    f2 = b * x0 / (x0 * x0 + b * x0 + c)
    f3 = 2 * (2 * x0 + b) / q
    rs12 = np.sqrt(rs)
    fx = rs + b * rs12 + c
    qx = np.arctan(q / (2 * rs12 + b))

    ec = a * (np.log(rs / fx) + f1 * qx - f2 * (np.log((rs12 - x0)**2 / fx) + f3 * qx))
    if exc_only:
        return ec, None

    tx = 2 * rs12 + b
    tt = tx * tx + q * q
    vc = ec - rs12 * a / 6 * (2 / rs12 - tx / fx - 4 * b / tt -
                              f2 * (2 / (rs12 - x0) - tx / fx - 4 * (2 * x0 + b) / tt))
    return ec, np.array([vc])


def lda_c_vwn_spin(n, zeta, exc_only=False, **kwargs):
    '''Vosko-Wilk-Nusair parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_VWN and ID 7 in Libxc.
    Reference: Phys. Rev. B 22, 3812.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: VWN correlation energy density and potential.
    '''
    pi34 = (3 / (4 * np.pi))**(1 / 3)
    rs = pi34 / n**(1 / 3)
    zeta4 = zeta**4

    cfz = 2**(4 / 3) - 2
    iddfz0 = 9 / 8 * cfz
    trup = 1 + zeta
    trdw = 1 - zeta
    trup13 = trup**(1 / 3)
    trdw13 = trdw**(1 / 3)

    def pade_fit(i):
        '''Calculate correlation energies by Pade approximation interpolation.

        Args:
            i (int): Index to choose paramagnetic (0), ferromagnetic (1), or spin stiffness (2) fit.

        Returns:
            tuple[ndarray, ndarray]: Pade fit and the derivative.
        '''
        a = (0.0310907, 0.01554535, -0.01688686394039)
        b = (3.72744, 7.06042, 1.13107)
        c = (12.9352, 18.0578, 13.0045)
        x0 = (-0.10498, -0.325, -0.0047584)
        Q = (6.15199081975908, 4.73092690956011, 7.12310891781812)
        tbQ = (1.21178334272806, 2.98479352354082, 0.31757762321188)
        bx0fx0 = (-0.03116760867894, -0.14460061018521, -0.00041403379428)

        sqrtrs = np.sqrt(rs)

        xx0 = sqrtrs - x0[i]
        Qtxb = Q[i] / (2 * sqrtrs + b[i])
        atg = np.arctan(Qtxb)
        fx = rs + b[i] * sqrtrs + c[i]
        fit = a[i] * (np.log(rs / fx) + tbQ[i] * atg -
                      bx0fx0[i] * (np.log(xx0 * xx0 / fx) + (tbQ[i] + 4 * x0[i] / Q[i]) * atg))
        if exc_only:
            return fit, None

        txb = 2 * sqrtrs + b[i]
        txbfx = txb / fx
        itxbQ = 1 / (txb * txb + Q[i] * Q[i])
        dfit = fit - a[i] / 3 + \
            a[i] * sqrtrs / 6 * (txbfx + 4 * b[i] * itxbQ + bx0fx0[i] *
                                 (2 / xx0 - txbfx - 4 * (b[i] + 2 * x0[i]) * itxbQ))
        return fit, dfit

    ecP, vcP = pade_fit(0)  # Paramagnetic fit
    ecF, vcF = pade_fit(1)  # Ferromagnetic fit
    ac, dac = pade_fit(2)   # Spin stiffness

    fz = (trup13 * trup + trdw13 * trdw - 2) / cfz  # f(zeta)
    ac *= iddfz0
    De = ecF - ecP - ac  # e_c[F] - e_c[P] - alpha_c/(ddf/ddz(z=0))
    fzz4 = fz * zeta4

    ec = ecP + ac * fz + De * fzz4
    if exc_only:
        return ec, None

    dfz = 4 / 3 * (trup13 - trdw13) / cfz  # df/dzeta
    dac *= iddfz0
    dec1 = vcP + dac * fz + (vcF - vcP - dac) * fzz4  # e_c-(r_s/3)*(de_c/dr_s)
    dec2 = ac * dfz + De * (4 * zeta**3 * fz + zeta4 * dfz)  # de_c/dzeta

    vcup = dec1 + (1 - zeta) * dec2
    vcdw = dec1 - (1 + zeta) * dec2
    return ec, np.array([vcup, vcdw])



#!/usr/bin/env python3
'''Calculate different energy contributions.'''
import numpy as np
from scipy.linalg import inv, norm
from scipy.special import erfc

from .dft import get_n_single, solve_poisson
from .xc import get_exc


class Energy:
    '''Energy class to save energy contributions in one place.'''
    __slots__ = ['Ekin', 'Eloc', 'Enonloc', 'Ecoul', 'Exc', 'Eewald', 'Esic']

    def __init__(self):
        self.Ekin = 0     # Kinetic energy
        self.Ecoul = 0    # Coulomb energy
        self.Exc = 0      # Exchange-correlation energy
        self.Eloc = 0     # Local energy
        self.Enonloc = 0  # Non-local energy
        self.Eewald = 0   # Ewald energy
        self.Esic = 0     # Self-interaction correction energy

    @property
    def Etot(self):
        '''Total energy is the sum of all energy contributions.'''
        return self.Ekin + self.Ecoul + self.Exc + self.Eloc + self.Enonloc + self.Eewald + \
            self.Esic

    def __repr__(self):
        '''Print the energies stored in the Energy object.'''
        out = ''
        for ie in self.__slots__:
            energy = getattr(self, ie)
            if energy != 0:
                out += f'{ie.ljust(8)}: {energy:+.9f} Eh\n'
        return f'{out}{"-" * 25}\nEtot    : {self.Etot:+.9f} Eh'


def get_E(scf):
    '''Calculate energy contributions and update energies needed in one SCF step.

    Args:
        scf: SCF object.

    Returns:
        float: Total energy.
    '''
    scf.energies.Ekin = get_Ekin(scf.atoms, scf.Y)
    scf.energies.Ecoul = get_Ecoul(scf.atoms, scf.n, scf.phi)
    scf.energies.Exc = get_Exc(scf, scf.n, scf.exc, scf.atoms.Nspin)
    scf.energies.Eloc = get_Eloc(scf, scf.n)
    scf.energies.Enonloc = get_Enonloc(scf, scf.Y)
    return scf.energies.Etot


def get_Ekin(atoms, Y):
    '''Calculate the kinetic energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        float: Kinetic energy in Hartree.
    '''
    # Ekin = -0.5 Tr(F Wdag L(W))
    Ekin = 0
    for spin in range(atoms.Nspin):
        F = np.diag(atoms.f[spin])
        Ekin += -0.5 * np.trace(F @ (Y[spin].conj().T @ atoms.L(Y[spin])))
    return np.real(Ekin)


def get_Ecoul(atoms, n, phi=None):
    '''Calculate the Coulomb energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        n (ndarray): Real-space electronic density.

    Keyword Args:
        phi (ndarray): Hartree ﬁeld.

    Returns:
        float: Coulomb energy in Hartree.
    '''
    if phi is None:
        phi = solve_poisson(atoms, n)
    # Ecoul = -(J(n))dag O(phi)
    return np.real(0.5 * n.T @ atoms.Jdag(atoms.O(phi)))


def get_Exc(scf, n, exc=None, n_spin=None, Nspin=2):
    '''Calculate the exchange-correlation energy.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        n (ndarray): Real-space electronic density.

    Keyword Args:
        exc (ndarray): Exchange-correlation energy density.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Returns:
        float: Exchange-correlation energy in Hartree.
    '''
    atoms = scf.atoms
    if exc is None:
        exc = get_exc(scf.xc, n_spin, Nspin)
    # Exc = (J(n))dag O(J(exc))
    return np.real(n.T @ atoms.Jdag(atoms.O(atoms.J(exc))))


def get_Eloc(scf, n):
    '''Calculate the local energy contribution.

    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        scf: SCF object.
        n (ndarray): Real-space electronic density.

    Returns:
        float: Local energy contribution in Hartree.
    '''
    return np.real(scf.Vloc.conj().T @ n)


# Adapted from https://github.com/f-fathurrahman/PWDFT.jl/blob/master/src/calc_energies.jl
def get_Enonloc(scf, Y):
    '''Calculate the non-local GTH energy contribution.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Returns:
        float: Non-local GTH energy contribution in Hartree.
    '''
    atoms = scf.atoms

    Enonloc = 0
    if scf.NbetaNL > 0:  # Only calculate non-local potential if necessary
        for spin in range(atoms.Nspin):
            betaNL_psi = (Y[spin].conj().T @ scf.betaNL).conj()

            enl = np.zeros(atoms.Nstate, dtype=complex)
            for ia in range(atoms.Natoms):
                psp = scf.GTH[atoms.atom[ia]]
                for l in range(psp['lmax']):
                    for m in range(-l, l + 1):
                        for iprj in range(psp['Nproj_l'][l]):
                            ibeta = scf.prj2beta[iprj, ia, l, m + psp['lmax'] - 1] - 1
                            for jprj in range(psp['Nproj_l'][l]):
                                jbeta = scf.prj2beta[jprj, ia, l, m + psp['lmax'] - 1] - 1
                                hij = psp['h'][l, iprj, jprj]
                                enl += hij * betaNL_psi[:, ibeta].conj() * betaNL_psi[:, jbeta]
            Enonloc += np.sum(atoms.f[spin] * enl)
    # We have to multiply with the cell volume, because of different orthogonalization methods
    return np.real(Enonloc * atoms.Omega)


def get_Eewald(atoms, gcut=2, gamma=1e-8):
    '''Calculate the Ewald energy.

    Reference: J. Chem. Theory Comput. 10, 381.

    Args:
        atoms: Atoms object.

    Keyword Args:
        gcut (float): G-vector cut-off.
        gamma (float): Error tolerance.

    Returns:
        float: Ewald energy in Hartree.
    '''
    # For a plane wave code we have multiple contributions for the Ewald energy
    # Namely, a sum from contributions from real-space, reciprocal space,
    # the self energy, (the dipole term [neglected]), and an additional electroneutrality term
    def get_index_vectors(s):
        '''Create all index vectors of periodic images.

        Args:
            s (ndarray): Number of images per lattice vector.

        Returns:
            ndarray: Index matrix.
        '''
        m1 = np.arange(-s[0], s[0] + 1)
        m2 = np.arange(-s[1], s[1] + 1)
        m3 = np.arange(-s[2], s[2] + 1)
        M = np.transpose(np.meshgrid(m1, m2, m3)).reshape(-1, 3)
        # Delete the [0, 0, 0] element, to prevent checking for it in every loop
        return M[~np.all(M == 0, axis=1)]

    # Calculate the Ewald splitting parameter nu
    gexp = -np.log(gamma)
    nu = 0.5 * np.sqrt(gcut**2 / gexp)

    # Start by calculating the self-energy
    Eewald = -nu / np.sqrt(np.pi) * np.sum(atoms.Z**2)
    # Add the electroneutrality term
    Eewald += -np.pi * np.sum(atoms.Z)**2 / (2 * nu**2 * atoms.Omega)

    # Calculate the real-space contribution
    # Calculate the amount of images that have to be considered per axis
    Rm = norm(atoms.R, axis=1)
    tmax = np.sqrt(0.5 * gexp) / nu
    s = np.rint(tmax / Rm + 1.5)
    # Collect all box index vector in a matrix
    M = get_index_vectors(s)
    # Calculate the translation vectors
    T = M @ atoms.R

    for ia in range(atoms.Natoms):
        for ja in range(atoms.Natoms):
            dX = atoms.X[ia] - atoms.X[ja]
            ZiZj = atoms.Z[ia] * atoms.Z[ja]
            rmag = np.sqrt(norm(dX - T, axis=1)**2)
            # Add the real-space contribution
            Eewald += 0.5 * ZiZj * np.sum(erfc(rmag * nu) / rmag)
            # The T=[0, 0, 0] element is ommited in M but needed if ia!=ja, so add it manually
            if ia != ja:
                rmag = np.sqrt(norm(dX)**2)
                Eewald += 0.5 * ZiZj * erfc(rmag * nu) / rmag

    # Calculate the reciprocal space contribution
    # Calculate the amount of reciprocal images that have to be considered per axis
    g = 2 * np.pi * inv(atoms.R.T)
    gm = norm(g, axis=1)
    s = np.rint(gcut / gm + 1.5)
    # Collect all box index vector in a matrix
    M = get_index_vectors(s)
    # Calculate the reciprocal translation vectors and precompute the prefactor
    G = M @ g
    G2 = norm(G, axis=1)**2
    prefactor = 2 * np.pi / atoms.Omega * np.exp(-0.25 * G2 / nu**2) / G2

    for ia in range(atoms.Natoms):
        for ja in range(atoms.Natoms):
            dX = atoms.X[ia] - atoms.X[ja]
            ZiZj = atoms.Z[ia] * atoms.Z[ja]
            GX = np.sum(G * dX, axis=1)
            # Add the reciprocal space contribution
            Eewald += ZiZj * np.sum(prefactor * np.cos(GX))
    return Eewald


def get_Esic(scf, Y, n_single=None):
    '''Calculate the Perdew-Zunger self-interaction energy.

    Reference: Phys. Rev. B 23, 5048.

    Args:
        scf: SCF object.
        Y (ndarray): Expansion coefficients of orthogonal wave functions in reciprocal space.

    Keyword Args:
        n_single (ndarray): Single-electron densities.

    Returns:
        float: PZ self-interaction energy.
    '''
    atoms = scf.atoms
    # E_PZ-SIC = \sum_i Ecoul[n_i] + Exc[n_i, 0]
    if n_single is None:
        n_single = get_n_single(atoms, Y)

    Esic = 0
    for i in range(atoms.Nstate):
        for spin in range(atoms.Nspin):
            if atoms.f[spin, i] > 0:
                # Create normalized single-particle densities in the form of electronic densities
                # per spin channel, since spin-polarized functionals expect this form
                ni = np.zeros((2, len(n_single[0])))
                # Normalize single-particle densities to 1
                ni[0] = n_single[spin, :, i] / atoms.f[spin, i]

                coul = get_Ecoul(atoms, ni[0])
                # The exchange part for a SIC correction has to be spin-polarized
                xc = get_Exc(scf, ni[0], n_spin=ni, Nspin=2)
                # SIC energy is scaled by the occupation number
                Esic += (coul + xc) * atoms.f[spin, i]
    scf.energies.Esic = Esic
    return Esic



#!/usr/bin/env python3
'''Functions to generate masks to restrict real-space fields to domains.'''
import numpy as np
from scipy.linalg import norm

from .tools import center_of_mass


def domain_cuboid(object, length, centers=None):
    '''Generate a mask for a cuboidal real-space domain.

    Args:
        object: Atoms or SCF object.
        length (int | float | list | ndarray): Side length or lengths of the cuboid.

    Keyword Args:
        centers (ndarray): Center of the cuboid.

            Defaults to the geometric center of mass of the system. For multiple coordinates,
            multiple domains will be merged.

    Returns:
        ndarray: Boolean mask.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if isinstance(length, (int, float)):
        length = length * np.ones(3)
    if centers is None:
        centers = center_of_mass(atoms.X)
    # Handle each dimension seperately and add them together
    if centers.ndim == 1:
        mask1 = (np.abs(centers[0] - atoms.r[:, 0]) < length[0])
        mask2 = (np.abs(centers[1] - atoms.r[:, 1]) < length[1])
        mask3 = (np.abs(centers[2] - atoms.r[:, 2]) < length[2])
        mask = mask1 & mask2 & mask3
    else:
        mask = np.zeros(len(atoms.r), dtype=bool)
        for center in centers:
            mask1 = (np.abs(center[0] - atoms.r[:, 0]) < length[0])
            mask2 = (np.abs(center[1] - atoms.r[:, 1]) < length[1])
            mask3 = (np.abs(center[2] - atoms.r[:, 2]) < length[2])
            mask = mask | (mask1 & mask2 & mask3)
    return mask


def domain_isovalue(field, isovalue):
    '''Generate a mask for an isovalue real-space domain.

    Args:
        field (ndarray): Real-space field data.
        isovalue (float): Isovalue for the truncation.

    Returns:
        ndarray: Boolean mask.
    '''
    return np.abs(field) > isovalue


def domain_sphere(object, radius, centers=None):
    '''Generate a mask for a spherical real-space domain.

    Args:
        object: Atoms or SCF object.
        radius (float): Radius of the sphere.

    Keyword Args:
        centers (ndarray): Center of the sphere.

            Defaults to the geometric center of mass of the system. For multiple coordinates,
            multiple domains will be merged.

    Returns:
        ndarray: Boolean mask.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if centers is None:
        centers = center_of_mass(atoms.X)
    if centers.ndim == 1:
        mask = norm(centers - atoms.r, axis=1) < radius
    else:
        mask = np.zeros(len(atoms.r), dtype=bool)
        for center in centers:
            mask_tmp = norm(center - atoms.r, axis=1) < radius
            mask = mask | mask_tmp
    return mask


def truncate(field, mask):
    '''Truncate field data for a given mask.

    This will not return a smaller array but set all truncated values to zero.

    Args:
        field (ndarray): Real-space field data.
        mask (ndarray): Boolean mask.

    Returns:
        ndarray: Boolean mask.
    '''
    field_trunc = np.copy(field)
    field_trunc[~mask] = 0
    return field_trunc



#!/usr/bin/env python3
'''Implementation of different exchange-correlation functionals.'''
from .utils import ALIAS, get_exc, get_vxc, get_xc, get_zeta, IMPLEMENTED, parse_functionals, XC_MAP

__all__ = ['ALIAS', 'get_exc', 'get_vxc', 'get_xc', 'get_zeta', 'IMPLEMENTED', 'parse_functionals',
           'XC_MAP']



#!/usr/bin/env python3
'''CUBE file handling.'''
import textwrap
import time

import numpy as np

from ..data import NUMBER2SYMBOL, SYMBOL2NUMBER
from ..logger import log
from ..version import __version__


def read_cube(filename):
    '''Load atom and cell data from cube files.

    There is no standard for cube files. The following format has been used.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        filename (str): cube input file path/name.

    Returns:
        tuple[list, ndarray, float, ndarray, int]: Species, positions, charges, cell size, sampling.
    '''
    if not filename.endswith('.cube'):
        filename += '.cube'

    # Atomic units and a cuboidal cell that starts at (0,0,0) are assumed.
    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first and second line can contain comments, print them if available
        comment = f'{lines[0].strip()}\n{lines[1].strip()}'
        log.info(f'XYZ file comment: "{comment}"')

        # Line 4 to 6 contain the sampling per axis, and the cell basis vectors with length a/s
        # A cuboidal cell is assumed, so only use the diagonal entries
        s = np.empty(3, dtype=int)
        a = np.empty(3)
        for i, line in enumerate(lines[3:6]):
            line_split = line.strip().split()
            s[i] = line_split[0]
            a[i] = s[i] * np.float_(line_split[i + 1])

        atom = []
        X = []
        Z = []
        # Following lines contain atom positions with the format: atom-id charge x-pos y-pos z-pos
        for line in lines[6:]:
            line_split = line.strip().split()
            # If the first value is not a (positive) integer, we have reached the field data
            if not line_split[0].isdigit():
                break
            atom.append(NUMBER2SYMBOL[int(line_split[0])])
            Z.append(line_split[1])
            X.append(np.float_(line_split[2:5]))

    X = np.asarray(X)
    return atom, X, Z, a, s


def write_cube(object, filename, field, fods=None, elec_symbols=None):
    '''Generate cube files from given field data.

    There is no standard for cube files. The following format has been used to work with VESTA.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        object: Atoms or SCF object.
        filename (str): cube output file path/name.
        field (ndarray): Real-space field data.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    '''
    # Atomic units are assumed, so there is no need for conversion.
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if not filename.endswith('.cube'):
        filename += '.cube'

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                        'polarized case.')

    # Our field data has been created in a different order than needed for cube files
    # (triple loop over z,y,x instead of x,y,z), so rearrange it with some index magic.
    idx = []
    for Nx in range(atoms.s[0]):
        for Ny in range(atoms.s[1]):
            for Nz in range(atoms.s[2]):
                idx.append(Nx + Ny * atoms.s[0] + Nz * atoms.s[0] * atoms.s[1])
    idx = np.asarray(idx)

    # Make sure we have real valued data in the correct order
    field = np.real(field[idx])

    with open(filename, 'w') as fp:
        # The first two lines have to be a comment.
        # Print information about the file and program, and the file creation time.
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n\n')
        # Number of atoms (int), and origin of the coordinate system (float)
        # The origin is normally at 0,0,0 but we could move our box, so take the minimum
        if fods is None:
            fp.write(f'{atoms.Natoms}  ')
        else:
            fp.write(f'{atoms.Natoms + len(fods[0]) + len(fods[1])}  ')
        fp.write('0.0  0.0  0.0\n')
        # Number of points per axis (int), and vector defining the axis (float)
        # We only have a cuboidal box, so each vector only has one non-zero component
        fp.write(f'{atoms.s[0]}  {atoms.a[0] / atoms.s[0]:.6f}  0.0  0.0\n'
                 f'{atoms.s[1]}  0.0  {atoms.a[1] / atoms.s[1]:.6f}  0.0\n'
                 f'{atoms.s[2]}  0.0  0.0  {atoms.a[2] / atoms.s[2]:.6f}\n')
        # Atomic number (int), atomic charge (float), and atom position (floats) for every atom
        for ia in range(atoms.Natoms):
            fp.write(f'{SYMBOL2NUMBER[atoms.atom[ia]]}  {atoms.Z[ia]:.3f}  '
                     f'{atoms.X[ia, 0]: .6f}  {atoms.X[ia, 1]: .6f}  {atoms.X[ia, 2]: .6f}\n')
        if fods is not None:
            for ie in fods[0]:
                fp.write(f'{SYMBOL2NUMBER[elec_symbols[0]]}  0.000  '
                         f'{ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
            for ie in fods[1]:
                fp.write(f'{SYMBOL2NUMBER[elec_symbols[1]]}  0.000  '
                         f'{ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
        # Field data (float) with scientific formatting
        # We have s[0]*s[1] chunks values with a length of s[2]
        for i in range(atoms.s[0] * atoms.s[1]):
            # Print every round of values, so we can add empty lines between them
            data_str = '%+1.6e  ' * atoms.s[2] % tuple(field[i * atoms.s[2]:(i + 1) * atoms.s[2]])
            # Print a maximum of 6 values per row
            # Max width for this formatting is 90, since 6*len('+1.00000e-000  ')=90
            fp.write(f'{textwrap.fill(data_str, width=90)}\n\n')
    return



#!/usr/bin/env python3
'''Modified Chachiyo LDA correlation.

Reference: Comput. Theor. Chem. 1172, 112669.
'''
from .lda_c_chachiyo import lda_c_chachiyo, lda_c_chachiyo_spin


def lda_c_chachiyo_mod(n, **kwargs):
    '''Modified Chachiyo parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_CHACHIYO_MOD and ID 307 in Libxc.
    Reference: Comput. Theor. Chem. 1172, 112669.

    Args:
        n (ndarray): Real-space electronic density.

    Returns:
        tuple[ndarray, ndarray]: Chachiyo correlation energy density and potential.
    '''
    return lda_c_chachiyo(n, **kwargs)


def chachiyo_mod_scaling(zeta, exc_only=False):
    '''Modified weighting factor between the paramagnetic and the ferromagnetic case.

    Reference: Comput. Theor. Chem. 1172, 112669.

    Args:
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: Weighting factor and its derivative.
    '''
    gzeta = ((1 + zeta)**(2 / 3) + (1 - zeta)**(2 / 3)) / 2
    fzeta = 2 * (1 - gzeta**3)
    if exc_only:
        return fzeta, None

    dfdzeta = -2 * gzeta**2 * (1 / (1 + zeta)**(1 / 3) - 1 / (1 - zeta)**(1 / 3))
    return fzeta, dfdzeta


def lda_c_chachiyo_mod_spin(n, zeta, **kwargs):
    '''Modified Chachiyo parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_CHACHIYO_MOD and ID 307 in Libxc.
    Reference: Comput. Theor. Chem. 1172, 112669.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Returns:
        tuple[ndarray, ndarray]: Chachiyo correlation energy density and potential.
    '''
    return lda_c_chachiyo_spin(n, zeta, weight_function=chachiyo_mod_scaling, **kwargs)



#!/usr/bin/env python3
'''XYZ file handling.'''
import time

import numpy as np

from ..logger import log
from ..units import ang2bohr, bohr2ang
from ..version import __version__


def read_xyz(filename):
    '''Load atom species and positions from xyz files.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        filename (str): xyz input file path/name.

    Returns:
        tuple[list, ndarray]: Atom species and positions.
    '''
    if not filename.endswith('.xyz'):
        filename += '.xyz'

    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first line contains the number of atoms
        Natoms = int(lines[0].strip())

        # The second line can contain a comment, print it if available
        comment = lines[1].strip()
        log.info(f'XYZ file comment: "{comment}"')

        atom = []
        X = []
        # Following lines contain atom positions with the format: Atom x-pos y-pos z-pos
        for line in lines[2:2 + Natoms]:
            line_split = line.strip().split()
            atom.append(line_split[0])
            X.append(np.float_(line_split[1:4]))

    # xyz files are in Angstrom, so convert to Bohr
    X = ang2bohr(np.asarray(X))
    return atom, X


def write_xyz(object, filename, fods=None, elec_symbols=None):
    '''Generate xyz files from atoms objects.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        object: Atoms or SCF object.
        filename (str): xyz output file path/name.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if not filename.endswith('.xyz'):
        filename += '.xyz'

    # Convert the coordinates from atomic units to Angstrom
    X = bohr2ang(atoms.X)
    if fods is not None:
        fods = [bohr2ang(i) for i in fods]

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                        'polarized case.')

    with open(filename, 'w') as fp:
        # The first line contains the number of atoms.
        # If we add FOD coordinates, add them to the count.
        if fods is None:
            fp.write(f'{atoms.Natoms}\n')
        else:
            fp.write(f'{atoms.Natoms + len(fods[0]) + len(fods[1])}\n')
        # The second line can contains a comment.
        # Print information about the file and program, and the file creation time.
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n')
        for ia in range(atoms.Natoms):
            fp.write(f'{atoms.atom[ia]:<2s}  {X[ia, 0]: .6f}  {X[ia, 1]: .6f}  {X[ia, 2]: .6f}\n')
        # Add FOD coordinates if desired. The atom symbol will default to X (no atom type).
        if fods is not None:
            for ie in fods[0]:
                fp.write(f'{elec_symbols[0]:<2s}  {ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
            for ie in fods[1]:
                fp.write(f'{elec_symbols[1]:<2s}  {ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
    return



#!/usr/bin/env python3
'''Interface to use Libxc functionals.

For a list of available functionals, see: https://www.tddft.org/programs/libxc/functionals

One can install pylibxc by compiling Libxc according to:
https://tddft.org/programs/libxc/installation/#python-library

Alternatively, one can use the PySCF Libxc interface with::

    pip install eminus[libxc]
'''
import numpy as np

from ..logger import log


def libxc_functional(xc, n_spin, Nspin):
    '''Handle Libxc exchange-correlation functionals via pylibxc.

    Only LDA functionals should be used.
    Reference: SoftwareX 7, 1.

    Args:
        xc (str | int): Exchange or correlation identifier.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Returns:
        tuple[ndarray, ndarray]: Exchange-correlation energy density and potential.
    '''
    try:
        from pylibxc import LibXCFunctional
    except ImportError:
        return pyscf_functional(xc, n_spin, Nspin)

    # Libxc functionals have one integer and one string identifier
    try:
        func = LibXCFunctional(int(xc), Nspin)
    except ValueError:
        func = LibXCFunctional(xc, Nspin)

    # Libxc expects a 1d array, so reshape n_spin (same as n_spin.ravel(order='F'))
    out = func.compute({'rho': n_spin.T.ravel()})
    # zk is a column vector, flatten it to a 1d row vector
    exc = out['zk'].ravel()
    # vrho is exactly transposed from what we need
    vxc = out['vrho'].T
    return exc, vxc


def pyscf_functional(xc, n_spin, Nspin):
    '''Handle Libxc exchange-correlation functionals via PySCF.

    Only LDA functionals should be used.
    Reference: WIREs Comput. Mol. Sci. 8, e1340.

    Args:
        xc (str | int): Exchange or correlation identifier.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Returns:
        tuple[ndarray, ndarray]: Exchange-correlation energy density and potential.
    '''
    try:
        from pyscf.dft.libxc import eval_xc
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[libxc]".\n\n')
        raise
    # The function also returns fxc and kxc (higher derivatives) that are not needed at this point
    # Spin in PySCF is the number of unpaired electrons, not the number of spin channels
    exc, vxc, _, _ = eval_xc(xc, n_spin, spin=Nspin - 1)
    # The first entry of vxc is vrho as a column array
    return exc, np.atleast_2d(vxc[0].T)



#!/usr/bin/env python3
'''Slater LDA exchange.

Reference: Phys. Rev. 81, 385.
'''
import numpy as np


def lda_x(n, exc_only=False, **kwargs):
    '''Slater exchange functional (spin-paired).

    Corresponds to the functional with the label LDA_X and ID 1 in Libxc.
    Reference: Phys. Rev. 81, 385.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: Exchange energy density and potential.
    '''
    pi34 = (3 / (4 * np.pi))**(1 / 3)
    f = -3 / 4 * (3 / (2 * np.pi))**(2 / 3)
    rs = pi34 / n**(1 / 3)

    ex = f / rs
    if exc_only:
        return ex, None

    vx = 4 / 3 * ex
    return ex, np.array([vx])


def lda_x_spin(n, zeta, exc_only=False, **kwargs):
    '''Slater exchange functional (spin-polarized).

    Corresponds to the functional with the label LDA_X and ID 1 in Libxc.
    Reference: Phys. Rev. 81, 385.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: Exchange energy density and potential.
    '''
    f = -3 / 4 * (3 / np.pi)**(1 / 3)

    rho13p = ((1 + zeta) * n)**(1 / 3)
    rho13m = ((1 - zeta) * n)**(1 / 3)

    exup = f * rho13p
    exdw = f * rho13m
    ex = 0.5 * ((1 + zeta) * exup + (1 - zeta) * exdw)
    if exc_only:
        return ex, None

    vxup = 4 / 3 * f * rho13p
    vxdw = 4 / 3 * f * rho13m
    return ex, np.array([vxup, vxdw])



#!/usr/bin/env python3
'''Minimization algorithms.'''
import logging

import numpy as np

from .dft import get_grad, get_n_spin, get_n_total, orth, solve_poisson
from .energies import get_E
from .logger import name
from .utils import dotprod
from .xc import get_xc


def scf_step(scf):
    '''Perform one SCF step for a DFT calculation.

    Calculating intermediate results speeds up the energy and gradient calculation.

    Args:
        scf: SCF object.

    Returns:
        float: Total energy.
    '''
    atoms = scf.atoms
    scf.Y = orth(atoms, scf.W)
    scf.n = get_n_total(atoms, scf.Y)
    scf.n_spin = get_n_spin(atoms, scf.Y, scf.n)
    scf.phi = solve_poisson(atoms, scf.n)
    scf.exc, scf.vxc = get_xc(scf.xc, scf.n_spin, atoms.Nspin)
    return get_E(scf)


def check_energies(scf, Elist, linmin='', cg=''):
    '''Check the energies for every SCF cycle and handle the output.

    Args:
        scf: SCF object.
        Elist (list): Total energies per SCF step.

    Keyword Args:
        linmin (float): Cosine between previous search direction and current gradient.
        cg (float): Conjugate-gradient orthogonality.

    Returns:
        bool: Convergence condition.
    '''
    iteration = len(Elist)

    if scf.log.level <= logging.DEBUG:
        with np.printoptions(formatter={'float': '{:+0.3e}'.format}):
            # Output handling
            if not isinstance(linmin, str):
                linmin = f' \tlinmin-test: {linmin}'
            if not isinstance(cg, str):
                cg = f' \tcg-test: {cg}'
            scf.log.debug(f'Iteration: {iteration} \tEtot: '
                          f'{scf.energies.Etot:+.{scf.print_precision}f}{linmin}{cg}')
    else:
        scf.log.info(f'Iteration: {iteration} \tEtot: {scf.energies.Etot:+.{scf.print_precision}f}')

    if iteration > 1:
        # Check for convergence
        if abs(Elist[-2] - Elist[-1]) < scf.etol:
            return True
        # Check if the current energy is lower than any of the last three values
        if np.any(np.asarray(Elist[-4:-1]) < Elist[-1]):
            scf.log.warning('Total energy is not decreasing.')
    return False


@name('steepest descent minimization')
def sd(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Steepest descent minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []

    for _ in range(Nit):
        c = cost(scf)
        costs.append(c)
        if condition(scf, costs):
            break
        for spin in range(atoms.Nspin):
            g = grad(scf, spin, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
            scf.W[spin] = scf.W[spin] - betat * g
    return costs


@name('line minimization')
def lm(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Line minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.Nspin)
    beta = np.empty(atoms.Nspin)

    # Gradients that need to be saved for each spin
    d = np.empty_like(scf.W, dtype=complex)

    # Do the first step without the linmin test
    for spin in range(atoms.Nspin):
        g = grad(scf, spin, scf.W)
        d[spin] = -g
        gt = grad(scf, spin, scf.W + betat * d[spin])
        beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
    # Update wave functions after calculating the gradients for each spin
    for spin in range(atoms.Nspin):
        scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        for spin in range(atoms.Nspin):
            g = grad(scf, spin, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
            # Calculate linmin each spin seperately
            if scf.log.level <= logging.DEBUG:
                linmin[spin] = dotprod(g, d[spin]) / \
                    np.sqrt(dotprod(g, g) * dotprod(d[spin], d[spin]))
            d[spin] = -g
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
        for spin in range(atoms.Nspin):
            scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin):
            break
    return costs


@name('preconditioned line minimization')
def pclm(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Preconditioned line minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.Nspin)
    beta = np.empty(atoms.Nspin)

    # Gradients that need to be saved for each spin
    d = np.empty_like(scf.W, dtype=complex)

    # Do the first step without the linmin test
    for spin in range(atoms.Nspin):
        g = grad(scf, spin, scf.W)
        d[spin] = -atoms.K(g)
        gt = grad(scf, spin, scf.W + betat * d[spin])
        beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
    # Update wave functions after calculating the gradients for each spin
    for spin in range(atoms.Nspin):
        scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        for spin in range(atoms.Nspin):
            g = grad(scf, spin, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
            # Calculate linmin each spin seperately
            if scf.log.level <= logging.DEBUG:
                linmin[spin] = dotprod(g, d[spin]) / \
                    np.sqrt(dotprod(g, g) * dotprod(d[spin], d[spin]))
            d[spin] = -atoms.K(g)
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
        for spin in range(atoms.Nspin):
            scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin):
            break
    return costs


@name('conjugate-gradient minimization')
def cg(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Conjugate-gradient minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.Nspin)
    cg = np.empty(atoms.Nspin)
    beta = np.empty(atoms.Nspin)

    # Gradients that need to be saved for each spin
    d = np.empty_like(scf.W, dtype=complex)
    d_old = np.empty_like(scf.W, dtype=complex)
    g_old = np.empty_like(scf.W, dtype=complex)

    # Do the first step without the linmin and cg test
    for spin in range(atoms.Nspin):
        g = grad(scf, spin, scf.W)
        d[spin] = -g
        gt = grad(scf, spin, scf.W + betat * d[spin])
        beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
        d_old[spin] = d[spin]
        g_old[spin] = g
    # Update wave functions after calculating the gradients for each spin
    for spin in range(atoms.Nspin):
        scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        for spin in range(atoms.Nspin):
            g = grad(scf, spin, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
            # Calculate linmin and cg for each spin seperately
            if scf.log.level <= logging.DEBUG:
                linmin[spin] = dotprod(g, d_old[spin]) / \
                    np.sqrt(dotprod(g, g) * dotprod(d_old[spin], d_old[spin]))
                cg[spin] = dotprod(g, g_old[spin]) / \
                    np.sqrt(dotprod(g, g) * dotprod(g_old[spin], g_old[spin]))
            if scf.cgform == 1:  # Fletcher-Reeves
                beta[spin] = dotprod(g, g) / dotprod(g_old[spin], g_old[spin])
            elif scf.cgform == 2:  # Polak-Ribiere
                beta[spin] = dotprod(g - g_old[spin], g) / \
                    dotprod(g_old[spin], g_old[spin])
            elif scf.cgform == 3:  # Hestenes-Stiefel
                beta[spin] = dotprod(g - g_old[spin], g) / \
                    dotprod(g - g_old[spin], d_old[spin])
            d[spin] = -g + beta[spin] * d_old[spin]
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
            d_old[spin] = d[spin]
            g_old[spin] = g
        for spin in range(atoms.Nspin):
            scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin, cg):
            break
    return costs


@name('preconditioned conjugate-gradient minimization')
def pccg(scf, Nit, cost=scf_step, grad=get_grad, condition=check_energies, betat=3e-5):
    '''Preconditioned conjugate-gradient minimization algorithm.

    Args:
        scf: SCF object.
        Nit (int): Maximum number of SCF steps.

    Keyword Args:
        cost (Callable): Function that will run every SCF step.
        grad (Callable): Function that calculates the respective gradient.
        condition (Callable): Function to check and log the convergence condition.
        betat (float): SCF step size.

    Returns:
        list: Total energies per SCF cycle.
    '''
    atoms = scf.atoms
    costs = []

    # Scalars that need to be saved for each spin
    linmin = np.empty(atoms.Nspin)
    cg = np.empty(atoms.Nspin)
    beta = np.empty(atoms.Nspin)

    # Gradients that need to be saved for each spin
    d = np.empty_like(scf.W, dtype=complex)
    d_old = np.empty_like(scf.W, dtype=complex)
    g_old = np.empty_like(scf.W, dtype=complex)

    # Do the first step without the linmin and cg test
    for spin in range(atoms.Nspin):
        g = grad(scf, spin, scf.W)
        d[spin] = -atoms.K(g)
        gt = grad(scf, spin, scf.W + betat * d[spin])
        beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
        d_old[spin] = d[spin]
        g_old[spin] = g
    # Update wave functions after calculating the gradients for each spin
    for spin in range(atoms.Nspin):
        scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

    c = cost(scf)
    costs.append(c)
    condition(scf, costs)

    for _ in range(1, Nit):
        for spin in range(atoms.Nspin):
            g = grad(scf, spin, scf.W, scf.Y, scf.n, scf.phi, scf.vxc)
            # Calculate linmin and cg for each spin seperately
            if scf.log.level <= logging.DEBUG:
                linmin[spin] = dotprod(g, d_old[spin]) / \
                    np.sqrt(dotprod(g, g) * dotprod(d_old[spin], d_old[spin]))
                cg[spin] = dotprod(g, atoms.K(g_old[spin])) / \
                    np.sqrt(dotprod(g, atoms.K(g)) * dotprod(g_old[spin], atoms.K(g_old[spin])))
            if scf.cgform == 1:  # Fletcher-Reeves
                beta[spin] = dotprod(g, atoms.K(g)) / dotprod(g_old[spin], atoms.K(g_old[spin]))
            elif scf.cgform == 2:  # Polak-Ribiere
                beta[spin] = dotprod(g - g_old[spin], atoms.K(g)) / \
                    dotprod(g_old[spin], atoms.K(g_old[spin]))
            elif scf.cgform == 3:  # Hestenes-Stiefel
                beta[spin] = dotprod(g - g_old[spin], atoms.K(g)) / \
                    dotprod(g - g_old[spin], d_old[spin])
            d[spin] = -atoms.K(g) + beta[spin] * d_old[spin]
            gt = grad(scf, spin, scf.W + betat * d[spin])
            beta[spin] = betat * dotprod(g, d[spin]) / dotprod(g - gt, d[spin])
            d_old[spin] = d[spin]
            g_old[spin] = g
        for spin in range(atoms.Nspin):
            scf.W[spin] = scf.W[spin] + beta[spin] * d[spin]

        c = cost(scf)
        costs.append(c)
        if condition(scf, costs, linmin, cg):
            break
    return costs


IMPLEMENTED = {
    'cg': cg,
    'lm': lm,
    'pccg': pccg,
    'pclm': pclm,
    'sd': sd
}



#!/usr/bin/env python3
'''PDB file handling.'''
import numpy as np

from ..logger import log
from ..units import bohr2ang


def write_pdb(object, filename, fods=None, elec_symbols=None):
    '''Generate pdb files from atoms objects.

    See :func:`~eminus.io.pdb.create_pdb_str` for more information about the pdb file format.

    Args:
        object: Atoms or SCF object.
        filename (str): pdb output file path/name.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if not filename.endswith('.pdb'):
        filename += '.pdb'

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                        'polarized case.')

    atom = atoms.atom
    X = atoms.X
    if fods is not None:
        atom += [elec_symbols[0]] * len(fods[0]) + [elec_symbols[1]] * len(fods[1])
        if fods[0].shape[0] != 0:
            X = np.vstack((X, fods[0]))
        if fods[1].shape[0] != 0:
            X = np.vstack((X, fods[1]))

    with open(filename, 'w') as fp:
        fp.write(create_pdb_str(atom, X, a=atoms.a))
    return


def create_pdb_str(atom, X, a=None):
    '''Convert atom symbols and positions to the pdb format.

    File format definitions:
        CRYST1: https://www.wwpdb.org/documentation/file-format-content/format33/sect8.html#CRYST1

        ATOM: https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Keyword Args:
        a (float): Cell size.

    Returns:
        str: pdb file format.
    '''
    # Convert Bohr to Angstrom
    X = bohr2ang(X)
    if a is not None:
        a = bohr2ang(a)

    # pdb files have specific numbers of characters for every data with changing justification
    # Write everything explicitly down to not loose track of line lengths
    pdb = ''
    # Create data for a cuboidal cell
    if a is not None:
        pdb += 'CRYST1'                 # 1-6 "CRYST1"
        pdb += f'{a[0]:9.3f}'.rjust(9)  # 7-15 a
        pdb += f'{a[1]:9.3f}'.rjust(9)  # 16-24 b
        pdb += f'{a[2]:9.3f}'.rjust(9)  # 25-33 c
        pdb += f'{90:7.2f}'.rjust(7)    # 34-40 alpha
        pdb += f'{90:7.2f}'.rjust(7)    # 41-47 beta
        pdb += f'{90:7.2f}'.rjust(7)    # 48-54 gamma
        pdb += ' '
        pdb += 'P 1'.ljust(11)          # 56-66 Space group
        # pdb += '1'.rjust(4)           # 67-70 Z value
        pdb += '\n'

    # Create molecule data
    pdb += 'MODEL 1'
    for ia in range(len(atom)):
        pdb += '\nATOM  '                   # 1-6 "ATOM"
        pdb += f'{ia + 1}'.rjust(5)         # 7-11 Atom serial number
        pdb += ' '
        pdb += f'{atom[ia]}'.rjust(4)       # 13-16 Atom name
        pdb += ' '                          # 17 Alternate location indicator
        pdb += 'MOL'                        # 18-20 Residue name
        pdb += ' '
        pdb += ' '                          # 22 Chain identifier
        pdb += '1'.rjust(4)                 # 23-26 Residue sequence number
        pdb += ' '                          # 27 Code for insertions of residues
        pdb += '   '
        pdb += f'{X[ia, 0]:8.3f}'.rjust(8)  # 31-38 X orthogonal coordinate
        pdb += f'{X[ia, 1]:8.3f}'.rjust(8)  # 39-46 Y orthogonal coordinate
        pdb += f'{X[ia, 2]:8.3f}'.rjust(8)  # 47-54 Z orthogonal coordinate
        pdb += f'{1:6.2f}'.rjust(6)         # 55-60 Occupancy
        pdb += f'{0:6.2f}'.rjust(6)         # 61-66 Temperature factor
        pdb += '          '
        pdb += f'{atom[ia]}'.rjust(2)       # 77-78 Element symbol
        # pdb += '  '                       # 79-80 Charge
    return f'{pdb}\nENDMDL'



#!/usr/bin/env python3
'''JSON file handling.'''
import base64
import json

import numpy as np


def read_json(filename):
    '''Load objects from a JSON file.

    Args:
        filename (str): json input file path/name.

    Returns:
        Class object.
    '''
    import eminus

    def custom_object_hook(dct):
        '''Custom JSON object hook to create eminus classes after deserialization.'''
        # ndarrays are base64 encoded, decode and recreate
        if isinstance(dct, dict) and '__ndarray__' in dct:
            data = base64.b64decode(dct['__ndarray__'])
            return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])

        # Create a simple eminus objects and set all attributes afterwards
        # Explicitly call objects with verbosity since the logger is created at instantiation
        if isinstance(dct, dict) and 'atom' in dct:
            atoms = eminus.Atoms(dct['atom'], dct['X'], verbose=dct['_verbose'])
            for attr in dct:
                if attr == 'log':
                    continue
                setattr(atoms, attr, dct[attr])
            # The tuple type is not preserved when serializing, manually cast the only important one
            if isinstance(atoms.active, list):
                atoms.active = tuple(atoms.active)
            return atoms
        if isinstance(dct, dict) and 'atoms' in dct:
            scf = eminus.SCF(dct['atoms'], verbose=dct['_verbose'])
            for attr in dct:
                if attr == 'log':
                    continue
                setattr(scf, attr, dct[attr])
            return scf
        if isinstance(dct, dict) and 'Ekin' in dct:
            energies = eminus.energies.Energy()
            for attr in dct:
                setattr(energies, attr, dct[attr])
            return energies
        return dct

    if not filename.endswith('.json'):
        filename += '.json'

    with open(filename, 'r') as fh:
        return json.load(fh, object_hook=custom_object_hook)


def write_json(object, filename):
    '''Save objects in a JSON file.

    Args:
        object: Class object.
        filename (str): json output file path/name.
    '''
    import eminus

    class CustomEncoder(json.JSONEncoder):
        '''Custom JSON encoder class to serialize eminus classes.'''
        def default(self, obj):
            # ndarrays are not json serializable, encode them as base64 to save them
            if isinstance(obj, np.ndarray):
                data = base64.b64encode(obj.copy(order='C')).decode('utf-8')
                return {'__ndarray__': data, 'dtype': str(obj.dtype), 'shape': obj.shape}

            # If obj is a Atoms or SCF class dump them as a dictionary
            if isinstance(obj, eminus.Atoms) or isinstance(obj, eminus.SCF):
                # Only dumping the dict would result in a string, so do one dump and one load
                data = json.dumps(obj.__dict__, cls=CustomEncoder)
                return dict(json.loads(data))
            # __slots__ classes have no __dict__, so use getattr
            if isinstance(obj, eminus.energies.Energy):
                data = json.dumps({slot: getattr(obj, slot) for slot in obj.__slots__})
                return dict(json.loads(data))
            # The logger class is not serializable, just ignore it
            if isinstance(obj, eminus.logger.CustomLogger):
                return None
            return json.JSONEncoder.default(self, obj)

    if not filename.endswith('.json'):
        filename += '.json'

    with open(filename, 'w') as fp:
        json.dump(object, fp, cls=CustomEncoder)
    return



#!/usr/bin/env python3
'''Package version number and version info function.'''
import importlib
import os
import platform
import sys

__version__ = '2.2.1'
logo = (' ___ _____ _ ___ _ _ ___ \n'
        '| -_|     | |   | | |_ -|\n'
        '|___|_|_|_|_|_|_|___|___|\n')


def info():
    '''Print version numbers and availability of packages.'''
    dependencies = ('numpy', 'scipy')
    extras = ('plotly', 'nglview', 'notebook', 'pyscf', 'pyflosic2')
    dev = ('pylibxc', 'pytest', 'flake8', 'sphinx', 'furo')

    print(logo)
    print('--- Platform infos ---'
          f'\nPlatform    : {platform.system()} {platform.machine()}'
          f'\nRelease     : {platform.release()} {platform.version()}'
          '\n\n--- Version infos ---'
          f'\npython      : {sys.version.split()[0]}'
          f'\neminus      : {__version__}')
    for pkg in dependencies + extras + dev:
        try:
            module = importlib.import_module(pkg)
            try:
                print(f'{pkg.ljust(12)}: {module.__version__}')
            except AttributeError:
                # pylibxc does not use the standard version identifier
                print(f'{pkg.ljust(12)}: {module.version.__version__}')
        except ModuleNotFoundError:
            if pkg in dependencies:
                print(f'{pkg.ljust(12)}: Dependency not installed')
            elif pkg in extras:
                print(f'{pkg.ljust(12)}: Extra not installed')

    print('\n--- Performance infos ---')
    try:
        THREADS = int(os.environ['OMP_NUM_THREADS'])
    except KeyError:
        THREADS = 1
        print('INFO: No OMP_NUM_THREADS environment variable was found.\n'
              'To improve performance, add "export OMP_NUM_THREADS=THREADS" to your ".bashrc".\n'
              'Make sure to replace "THREADS", typically with the number of cores your CPU has.')
    finally:
        print(f'FFT operations will run on {THREADS} thread{"s" if THREADS != 1 else ""}.')
    return


if __name__ == '__main__':
    info()



#!/usr/bin/env python3
'''File input and output functionalities.'''
from .cube import read_cube, write_cube
from .gth import read_gth
from .json import read_json, write_json
from .pdb import create_pdb_str, write_pdb
from .xyz import read_xyz, write_xyz
from ..logger import log

__all__ = ['create_pdb_str', 'read', 'read_cube', 'read_gth', 'read_json', 'read_xyz', 'write',
           'write_cube', 'write_json', 'write_pdb', 'write_xyz']


def read(*args, **kwargs):
    '''Unified file reader function.'''
    if args[0].endswith('.json'):
        return read_json(*args, **kwargs)
    elif args[0].endswith('.xyz'):
        return read_xyz(*args, **kwargs)
    elif args[0].endswith('.cube'):
        return read_cube(*args, **kwargs)
    else:
        log.error('No viable file ending found.')


def write(*args, **kwargs):
    '''Unified file writer function.'''
    if args[1].endswith('.json'):
        return write_json(*args, **kwargs)
    elif args[1].endswith('.xyz'):
        return write_xyz(*args, **kwargs)
    elif args[1].endswith('.cube'):
        return write_cube(*args, **kwargs)
    elif args[1].endswith('.pdb'):
        return write_pdb(*args, **kwargs)
    else:
        log.error('No viable file ending found.')



#!/usr/bin/env python3
'''Atoms class definition.'''
import re

import numpy as np
from scipy.fft import next_fast_len
from scipy.linalg import det, eig, inv, norm

from .io import read_gth
from .logger import create_logger, get_level, log
from .operators import I, Idag, J, Jdag, K, L, Linv, O, T
from .tools import center_of_mass, cutoff2gridspacing, inertia_tensor


class Atoms:
    '''Atoms object that holds all system and cell parameters.

    Args:
        atom (str | list | tuple): Atom symbols.

            Examples: 'CH4'; ['C', 'H', 'H', 'H', 'H']; ('C', 'H', 'H', 'H', 'H')
        X (list | tuple | ndarray): Atom positions.

            Examples: (0, 0, 0); array([0, 0, 0]); [[0, 0, 0], [1, 1, 1]];

    Keyword Args:
        a (float | list | tuple | ndarray): Cell size or vacuum size.

            A cuboidal box with the same side lengths will be created.

            Examples: 10; [10, 10, 10]; (7, 8, 9),
            Default: 20 Bohr (ca. 10.5 Angstrom).
        ecut (float | None): Cut-off energy.

            None will disable the G-Vector truncation (needs a separate s).
            Default: 30 Hartree (ca. 816 eV).
        Z (int | list | tuple | ndarray | None): Valence charge per atom.

            The charges should not differ for same species. None will use valence charges from GTH
            files. The same charge for every atom will be assumed for single integers.

            Example: 1; [4, 1, 1, 1, 1],
            Default: None
        s (int | list | tuple | ndarray | None): Real-space sampling of the cell.

            None will make the sampling dependent on a and ecut.

            Example: 30; [30, 40, 50]; array([30, 40, 50]),
            Default: None
        center (bool | str): Center the system inside the cell (case insensitive).

            Align the geometric center of mass with the center of the call and rotate the system,
            such that its geometric moment of inertia aligns with the coordinate axes.

            Example: True; 'shift'; 'rotate',
            Default: False
        Nspin (int | None): Number of spin states.

            1 for spin-paired, 2 for spin-polarized, None for automatic detection.
            Default: 2
        f (float | list | tuple | ndarray | None): Occupation numbers per state.

            The last state will be adjusted if the sum of f is not equal to the sum of Z.
            None will assume occupations of 2.

            Example: 2; [2, 2, 2, 2]; array([2, 2/3, 2/3, 2/3]),
            Default: None
        Nstate (int | None): Number of states.

            None will get the number of states from f or assume occupations of 2 and divide the sum
            of Z by it.

            Default: None
        verbose (int | str | None): Level of output (case insensitive).

            Can be one of 'CRITICAL', 'ERROR', 'WARNING', 'INFO', or 'DEBUG'.
            An integer value can be used as well, where larger numbers mean more output, starting
            from 0.
            None will use the default global logger value 'WARNING'.

            Default: 'info'
    '''
    def __init__(self, atom, X, a=20, ecut=30, Z=None, s=None, center=False, Nspin=2, f=None,
                 Nstate=None, verbose='info'):
        self.atom = atom      # Atom symbols
        self.X = X            # Atom positions
        self.a = a            # Cell/Vacuum size
        self.ecut = ecut      # Cut-off energy
        self.Z = Z            # Valence charges
        self.s = s            # Cell sampling
        self.center = center  # Center molecule in cell
        self.Nspin = Nspin    # Number of spin states
        self.f = f            # Occupation numbers
        self.Nstate = Nstate  # Number of states

        # Set up parameters that should not be cleared
        self.Natoms = None     # Number of atoms
        self.R = None          # Cell
        self.Omega = None      # Cell volume
        self.clear()

        # Initialize logger and update
        self.log = create_logger(self)
        if verbose is None:
            self.verbose = log.verbose
        else:
            self.verbose = verbose
        self.initialize()

    def clear(self):
        '''Initialize and clear parameters that will be built out of the inputs.'''
        self.r = None          # Sample points in cell
        self.G = None          # G-vectors
        self.G2 = None         # Squared magnitudes of G-vectors
        self.active = None     # Mask for active G-vectors
        self.G2c = None        # Truncated squared magnitudes of G-vectors
        self.Sf = None         # Structure factor
        self.is_built = False  # Flag to determine if the object was built or not
        return self

    def initialize(self):
        '''Validate inputs and update them if necessary.'''
        self._set_atom()
        self._set_charge()
        self._set_cell_size()
        self._set_positions()
        self._set_sampling()
        self._set_states(self.Nspin)
        return self

    def build(self):
        '''Build all necessary parameters.'''
        M, N = self._get_index_matrices()
        self._set_cell(M)
        self._set_G(N)
        self.is_built = True
        return self

    kernel = build

    def recenter(self, center=None):
        '''Recenter the system inside the cell.

        Keyword Args:
            center (float | list | tuple | ndarray | None): Point to center the system around.
        '''
        com = center_of_mass(self.X)
        if center is None:
            self.X = self.X - (com - self.a / 2)
        else:
            center = np.asarray(center)
            self.X = self.X - (com - center)

        # Recalculate the structure factor since it depends on the atom positions
        _, N = self._get_index_matrices()
        self._set_G(N)
        return self

    def _set_atom(self):
        '''Validate the atom input and calculate the number of atoms.'''
        # Quick option to set the charge for single atoms
        if isinstance(self.atom, str) and '-q' in self.atom:
            atom, Z = self.atom.split('-q')
            self.atom = [atom]
            self.Z = np.asarray(list(Z), dtype=int)
        # If a string is given for atom symbols convert them to a list of strings
        if isinstance(self.atom, str):
            # Insert a whitespace before every capital letter, these can appear once or none at all
            # Or insert before digits, these can appear at least once
            self.atom = re.sub(r'([A-Z?]|\d+)', r' \1', self.atom).split()
            atom = []
            for ia in self.atom:
                if ia.isdigit():
                    # if ia is a integer, append the previous atom ia-1 times
                    atom += [atom[-1]] * (int(ia) - 1)
                else:
                    # If ia is a string, add it to the results list
                    atom += [ia]
            self.atom = atom

        # Get the number of atoms
        self.Natoms = len(self.atom)
        return

    def _set_charge(self):
        '''Validate the Z input and calculate charges if necessary.'''
        # If only one charge is given, assume it is the charge for every atom
        if isinstance(self.Z, (int, np.integer)):
            self.Z = [self.Z] * self.Natoms
        if isinstance(self.Z, (list, tuple)):
            self.Z = np.asarray(self.Z)

        # If no charge is given, use the ionic charge from the GTH files
        if self.Z is None:
            Z = []
            for ia in range(self.Natoms):
                gth_dict = read_gth(self.atom[ia])
                Z.append(gth_dict['Zion'])
            self.Z = np.asarray(Z)
        return

    def _set_cell_size(self):
        '''Validate the a input.'''
        # Do this early on, since it is needed in many functions
        if isinstance(self.a, (int, np.integer, float, np.floating)):
            self.a = self.a * np.ones(3)
        if isinstance(self.a, (list, tuple)):
            self.a = np.asarray(self.a)
        return

    def _set_positions(self):
        '''Validate the X and center input and center the system if desired.'''
        # We need atom positions as an two-dimensional array
        self.X = np.atleast_2d(self.X)
        if isinstance(self.center, str):
            self.center = self.center.lower()

        # Center system such that the geometric inertia tensor will be diagonal
        # Rotate before shifting!
        if self.center or self.center == 'rotate':
            X = self.X
            I = inertia_tensor(self.X)
            _, eigvecs = eig(I)
            self.X = (inv(eigvecs) @ self.X.T).T

        # Shift system such that its geometric center of mass is in the center of the cell
        if self.center or self.center == 'shift':
            X = self.X
            com = center_of_mass(X)
            self.X = X - (com - self.a / 2)
        return

    def _set_sampling(self):
        '''Validate the s input and calculate it if necessary.'''
        # Make sampling dependent of ecut if no sampling is given
        if self.s is None:
            try:
                s = np.int_(self.a / cutoff2gridspacing(self.ecut))
            except TypeError:
                self.log.exception('No ecut provided, please enter a valid s.')
                raise
            # Multiply by two and add one to match PWDFT
            s = 2 * s + 1
            # Calculate a fast length to optimize the FFT calculations
            # See https://github.com/scipy/scipy/blob/main/scipy/fft/_helper.py
            self.s = [next_fast_len(i) for i in s]

        # Choose the same sampling for every direction if an integer is given
        if isinstance(self.s, (int, np.integer)):
            self.s = self.s * np.ones(3, dtype=int)
        if isinstance(self.s, (list, tuple)):
            self.s = np.asarray(self.s)
        return

    def _set_states(self, Nspin):
        '''Validate the f and Nstate input and calculate the states if necessary.

        Args:
            Nspin (int | None): Number of spin states.
        '''
        # Use a spin-paired calculation for an even number of electrons
        if Nspin is None:
            if np.sum(self.Z) % 2 == 0:
                Nspin = 1
            else:
                Nspin = 2
        # Make sure Nspin is an integer
        try:
            self.Nspin = int(Nspin)
        except (TypeError, ValueError):
            self.log.exception('Nspin has to be an integer.')
            raise
        # Make sure the occupations are in an two-dimensional array
        if isinstance(self.f, (list, tuple)):
            self.f = np.atleast_2d(self.f)
        # If occupations and spin number is not equal, reset the occupations and number of states
        if isinstance(self.f, np.ndarray) and len(self.f) != self.Nspin:
            self.f = None
            self.Nstate = None

        # If no states are provided use the length of the occupations
        if isinstance(self.f, np.ndarray) and self.Nstate is None:
            self.Nstate = len(self.f[0])
        # If one occupation and the number of states is given, use it for every state
        if isinstance(self.f, (int, np.integer, float, np.floating)) and self.Nstate is not None:
            self.f = self.f * np.ones((self.Nspin, self.Nstate))
        # If no occupations and the number of states is given, assume 1 or 2
        if self.f is None and self.Nstate is not None:
            self.f = 2 / self.Nspin * np.ones((self.Nspin, self.Nstate))

        # If the number of states is None and the occupations is a number or None, we are in trouble
        if self.Nstate is None:
            # If no occupations is given, assume 1 or 2
            if self.f is None:
                f = 2 / self.Nspin
            # Assume the number of states by dividing the total valence charge by an occupation of 2
            Ztot = np.sum(self.Z)
            self.Nstate = int(np.ceil(Ztot / 2))
            self.f = f * np.ones((self.Nspin, self.Nstate))
            # Subtract the leftovers from the last spin state
            self.f[-1, -1] -= np.sum(self.Z) % 2
        return

    def _get_index_matrices(self):
        '''Build index matrices M and N to build the real and reciprocal space samplings.

        Returns:
            tuple[ndarray, ndarray]: Index matrices.
        '''
        # Build index matrix M
        ms = np.arange(np.prod(self.s))
        m1 = ms % self.s[0]
        m2 = np.floor(ms / self.s[0]) % self.s[1]
        m3 = np.floor(ms / (self.s[0] * self.s[1])) % self.s[2]
        M = np.column_stack((m1, m2, m3))

        # Build index matrix N
        n1 = m1 - (m1 > self.s[0] / 2) * self.s[0]
        n2 = m2 - (m2 > self.s[1] / 2) * self.s[1]
        n3 = m3 - (m3 > self.s[2] / 2) * self.s[2]
        N = np.column_stack((n1, n2, n3))
        return M, N

    def _set_cell(self, M):
        '''Build cell and create the respective sampling.

        Args:
            M (ndarray): Index matrix.
        '''
        # Build a cuboidal cell and calculate its volume
        R = self.a * np.eye(3)
        self.R = R
        self.Omega = np.abs(det(R))
        # Build real-space sampling points
        self.r = M @ inv(np.diag(self.s)) @ R.T
        return

    def _set_G(self, N):
        '''Build G-vectors, build squared magnitudes G2, and generate the active space.

        Args:
            N (ndarray): Index matrix.
        '''
        # Build G-vectors
        G = 2 * np.pi * N @ inv(self.R)
        self.G = G
        # Calculate squared-magnitudes of G-vectors
        G2 = norm(G, axis=1)**2
        self.G2 = G2

        # Calculate the G2 restriction
        if self.ecut is not None:
            active = np.nonzero(G2 <= 2 * self.ecut)
        else:
            active = np.nonzero(G2 >= 0)  # Trivial condition to produce the right shape
        self.active = active
        self.G2c = G2[active]

        # Calculate the structure factor per atom
        self.Sf = np.exp(1j * G @ self.X.T).T
        return

    def __repr__(self):
        '''Print the parameters stored in the Atoms object.'''
        out = 'Atom\tCharge\tPosition'
        for i in range(self.Natoms):
            out += f'\n{self.atom[i]}\t{self.Z[i]}\t' \
                   f'{self.X[i, 0]:10.5f}  {self.X[i, 1]:10.5f}  {self.X[i, 2]:10.5f}'
        return out

    @property
    def verbose(self):
        '''Verbosity level.'''
        return self._verbose

    @verbose.setter
    def verbose(self, level):
        '''Verbosity setter to sync the logger with the property.'''
        self._verbose = get_level(level)
        self.log.setLevel(self._verbose)
        return

    def O(self, inp):
        '''Overlap operator :func:`~eminus.operators.O`.'''
        return O(self, inp)

    def L(self, inp):
        '''Laplacian operator :func:`~eminus.operators.L`.'''
        return L(self, inp)

    def Linv(self, inp):
        '''Inverse Laplacian operator :func:`~eminus.operators.Linv`.'''
        return Linv(self, inp)

    def I(self, inp):
        '''Transformation from reciprocal to real-space :func:`~eminus.operators.I`.'''
        return I(self, inp)

    def J(self, inp, full=True):
        '''Transformation from real to reciprocal space :func:`~eminus.operators.J`.'''
        return J(self, inp, full)

    def Idag(self, inp, full=False):
        '''Conj transformation from real to reciprocal space :func:`~eminus.operators.Idag`.'''
        return Idag(self, inp, full)

    def Jdag(self, inp):
        '''Conj transformation from reciprocal to real-space :func:`~eminus.operators.Jdag`.'''
        return Jdag(self, inp)

    def K(self, inp):
        '''Preconditioning operator :func:`~eminus.operators.K`.'''
        return K(self, inp)

    def T(self, inp, dr):
        '''Translation operator :func:`~eminus.operators.T`.'''
        return T(self, inp, dr)



#!/usr/bin/env python3
'''Linear algebra calculation utilities.'''
import functools

import numpy as np
from scipy.linalg import norm

from .logger import log


def diagprod(a, B):
    '''Efficiently calculate the expression Diag(a) * B.

    Args:
        a (ndarray): Single vector.
        B (ndarray): Matrix.

    Returns:
        ndarray: The expressions result.
    '''
    # Reshape a to a column vector
    a_col = a[:, None]
    return a_col * B


def dotprod(a, b):
    '''Efficiently calculate the expression a * b.

    Add an extra check to make sure the result is never zero since this function is used as a
    denominator in minimizers.

    Args:
        a (ndarray): Array of vectors.
        b (ndarray): Array of vectors.

    Returns:
        float: The expressions result
    '''
    eps = 1e-15  # 2.22e-16 is the range of float64 machine precision
    res = np.real(np.trace(a.conj().T @ b))
    if abs(res) < eps:
        return eps
    return res


def Ylm_real(l, m, G):
    '''Calculate real spherical harmonics from cartesian coordinates.

    Args:
        l (int): Angular momentum number.
        m (int): Magnetic quantum number.
        G (ndarray): Recipocal lattice vector or array of lattice vectors.

    Returns:
        ndarray: Real spherical harmonics.
    '''
    eps = 1e-9
    # Account for single vectors
    G = np.atleast_2d(G)

    # No need to calculate more for l=0
    if l == 0:
        return 0.5 * np.sqrt(1 / np.pi) * np.ones(len(G))

    # cos(theta)=Gz/|G|
    Gm = norm(G, axis=1)
    with np.errstate(divide='ignore', invalid='ignore'):
        cos_theta = G[:, 2] / Gm
    # Account for small magnitudes, if norm(G) < eps: cos_theta=0
    cos_theta[Gm < eps] = 0

    # Vectorized version of sin(theta)=sqrt(max(0, 1-cos_theta^2))
    sin_theta = np.sqrt(np.amax((np.zeros_like(cos_theta), 1 - cos_theta**2), axis=0))

    # phi=arctan(Gy/Gx)
    phi = np.arctan2(G[:, 1], G[:, 0])
    # If Gx=0: phi=pi/2*sign(Gy)
    phi_idx = (eps > G[:, 0]) & (G[:, 0] > -eps)
    phi[phi_idx] = np.pi / 2 * np.sign(G[phi_idx, 1])

    if l == 1:
        if m == -1:   # py
            return 0.5 * np.sqrt(3 / np.pi) * sin_theta * np.sin(phi)
        elif m == 0:  # pz
            return 0.5 * np.sqrt(3 / np.pi) * cos_theta
        elif m == 1:  # px
            return 0.5 * np.sqrt(3 / np.pi) * sin_theta * np.cos(phi)
    elif l == 2:
        if m == -2:    # dxy
            return np.sqrt(15 / 16 / np.pi) * sin_theta**2 * np.sin(2 * phi)
        elif m == -1:  # dyz
            return np.sqrt(15 / 4 / np.pi) * cos_theta * sin_theta * np.sin(phi)
        elif m == 0:   # dz2
            return 0.25 * np.sqrt(5 / np.pi) * (3 * cos_theta**2 - 1)
        elif m == 1:   # dxz
            return np.sqrt(15 / 4 / np.pi) * cos_theta * sin_theta * np.cos(phi)
        elif m == 2:   # dx2-y2
            return np.sqrt(15 / 16 / np.pi) * sin_theta**2 * np.cos(2 * phi)
    elif l == 3:
        if m == -3:
            return 0.25 * np.sqrt(35 / 2 / np.pi) * sin_theta**3 * np.sin(3 * phi)
        elif m == -2:
            return 0.25 * np.sqrt(105 / np.pi) * sin_theta**2 * cos_theta * np.sin(2 * phi)
        elif m == -1:
            return 0.25 * np.sqrt(21 / 2 / np.pi) * sin_theta * (5 * cos_theta**2 - 1) * np.sin(phi)
        elif m == 0:
            return 0.25 * np.sqrt(7 / np.pi) * (5 * cos_theta**3 - 3 * cos_theta)
        elif m == 1:
            return 0.25 * np.sqrt(21 / 2 / np.pi) * sin_theta * (5 * cos_theta**2 - 1) * np.cos(phi)
        elif m == 2:
            return 0.25 * np.sqrt(105 / np.pi) * sin_theta**2 * cos_theta * np.cos(2 * phi)
        elif m == 3:
            return 0.25 * np.sqrt(35 / 2 / np.pi) * sin_theta**3 * np.cos(3 * phi)

    log.error(f'No definition found for Ylm({l}, {m})')
    return


def handle_spin_gracefully(func):
    '''Handle spin calculating the function for each channel seperately.

    This can only be applied if the only spin-dependent indexing is the wave function W.

    Implementing the explicit handling of spin adds an extra layer of complexity where one has to
    loop over the spin states in many places. We can hide this complexity using this decorator while
    still supporting many usecases, e.g., the operators previously act on arrays containing wave
    functions of all states and of one state only. This decorator maintains this functionality and
    adds the option to act on arrays containing wave functions of all spins and all states as well.

    Args:
        func (Callable): Function that acts on spin-states.

    Returns:
        Callable: Decorator.
    '''
    @functools.wraps(func)
    def decorator(object, W, *args, **kwargs):
        if W.ndim == 3:
            # If one is brave enough one could add multiprocessing over spin states right here
            return np.asarray([func(object, Wspin, *args, **kwargs) for Wspin in W])
        return func(object, W, *args, **kwargs)
    return decorator


def pseudo_uniform(size, seed=1234):
    '''Lehmer random number generator, following MINSTD.

    Reference: Commun. ACM. 12, 85.

    Args:
        size (tuple): Dimension of the array to create.

    Keyword Args:
        seed (int): Seed to initialize the random number generator.

    Returns:
        ndarray: Array with (pseudo) random numbers.
    '''
    W = np.zeros(size, dtype=complex)
    mult = 48271
    mod = (2**31) - 1
    x = (seed * mult + 1) % mod
    for i in range(size[0]):
        for j in range(size[1]):
            for k in range(size[2]):
                x = (x * mult + 1) % mod
                W[i, j, k] = x / mod
    return W



#!/usr/bin/env python3
'''Perdew-Wang LDA correlation.

Reference: Phys. Rev. B 45, 13244.
'''
import numpy as np


def lda_c_pw(n, a=0.031091, exc_only=False, **kwargs):
    '''Perdew-Wang parametrization of the correlation functional (spin-paired).

    Corresponds to the functional with the label LDA_C_PW and ID 12 in Libxc.
    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.

    Keyword Args:
        a (float): Functional parameter.
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    '''
    a1 = 0.2137
    b1 = 7.5957
    b2 = 3.5876
    b3 = 1.6382
    b4 = 0.49294

    pi34 = (3 / (4 * np.pi))**(1 / 3)
    rs = pi34 / n**(1 / 3)
    rs12 = np.sqrt(rs)
    rs32 = rs * rs12
    rs2 = rs**2

    om = 2 * a * (b1 * rs12 + b2 * rs + b3 * rs32 + b4 * rs2)
    olog = np.log(1 + 1 / om)

    ec = -2 * a * (1 + a1 * rs) * olog
    if exc_only:
        return ec, None

    dom = 2 * a * (0.5 * b1 * rs12 + b2 * rs + 1.5 * b3 * rs32 + 2 * b4 * rs2)
    vc = -2 * a * (1 + 2 / 3 * a1 * rs) * olog - 2 / 3 * a * (1 + a1 * rs) * dom / (om * (om + 1))
    return ec, np.array([vc])


def lda_c_pw_spin(n, zeta, a=(0.031091, 0.015545, 0.016887), fz0=1.709921, exc_only=False,
                  **kwargs):
    '''Perdew-Wang parametrization of the correlation functional (spin-polarized).

    Corresponds to the functional with the label LDA_C_PW and ID 12 in Libxc.
    Reference: Phys. Rev. B 45, 13244.

    Args:
        n (ndarray): Real-space electronic density.
        zeta (ndarray): Relative spin polarization.

    Keyword Args:
        a (tuple): Functional parameters.
        fz0 (float): Functional parameter.
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: PW correlation energy density and potential.
    '''
    pi34 = (3 / (4 * np.pi))**(1 / 3)
    rs = pi34 / n**(1 / 3)
    zeta3 = zeta**3
    zeta4 = zeta3 * zeta

    def pw_fit(i):
        '''Calculate correlation energies by Perdew-Wang approximation interpolation.

        Args:
            i (int): Index to choose unpolarized (0), polarized (1), or antiferromagnetic (2) fit.

        Returns:
            tuple[ndarray, ndarray]: PW fit and the derivative.
        '''
        a1 = (0.2137, 0.20548, 0.11125)
        b1 = (7.5957, 14.1189, 10.357)
        b2 = (3.5876, 6.1977, 3.6231)
        b3 = (1.6382, 3.3662, 0.88026)
        b4 = (0.49294, 0.62517, 0.49671)

        rs12 = np.sqrt(rs)
        rs32 = rs * rs12
        rs2 = rs**2

        om = 2 * a[i] * (b1[i] * rs12 + b2[i] * rs + b3[i] * rs32 + b4[i] * rs2)
        olog = np.log(1 + 1 / om)
        fit = -2 * a[i] * (1 + a1[i] * rs) * olog
        if exc_only:
            return fit, None

        dom = 2 * a[i] * (0.5 * b1[i] * rs12 + b2[i] * rs + 1.5 * b3[i] * rs32 + 2 * b4[i] * rs2)
        dfit = -2 * a[i] * (1 + 2 / 3 * a1[i] * rs) * olog - \
            2 / 3 * a[i] * (1 + a1[i] * rs) * dom / (om * (om + 1))
        return fit, dfit

    ecU, vcU = pw_fit(0)  # Unpolarized
    ecP, vcP = pw_fit(1)  # Polarized
    ac, dac = pw_fit(2)   # Spin stiffness

    ac = -ac  # The PW spin interpolation parametrizes -ac instead of ac
    fz = ((1 + zeta)**(4 / 3) + (1 - zeta)**(4 / 3) - 2) / (2**(4 / 3) - 2)

    ec = ecU + ac * fz * (1 - zeta4) / fz0 + (ecP - ecU) * fz * zeta4
    if exc_only:
        return ec, None

    dac = -dac  # Also change the sign for the derivative
    dfz = ((1 + zeta)**(1 / 3) - (1 - zeta)**(1 / 3)) * 4 / (3 * (2**(4 / 3) - 2))
    factor1 = vcU + dac * fz * (1 - zeta4) / fz0 + (vcP - vcU) * fz * zeta4
    factor2 = (ac / fz0 * (dfz * (1 - zeta4) - 4 * fz * zeta3) +
               (ecP - ecU) * (dfz * zeta4 + 4 * fz * zeta3))

    vcup = factor1 + factor2 * (1 - zeta)
    vcdw = factor1 - factor2 * (1 + zeta)
    return ec, np.array([vcup, vcdw])



#!/usr/bin/env python3
'''Utility functions for exchange-correlation functionals.'''
import numpy as np

from .lda_c_chachiyo import lda_c_chachiyo, lda_c_chachiyo_spin
from .lda_c_chachiyo_mod import lda_c_chachiyo_mod, lda_c_chachiyo_mod_spin
from .lda_c_pw import lda_c_pw, lda_c_pw_spin
from .lda_c_pw_mod import lda_c_pw_mod, lda_c_pw_mod_spin
from .lda_c_vwn import lda_c_vwn, lda_c_vwn_spin
from .lda_x import lda_x, lda_x_spin
from ..logger import log


def get_xc(xc, n_spin, Nspin, dens_threshold=0, exc_only=False):
    '''Handle and get exchange-correlation functionals.

    Args:
        xc (list | str): Exchange and correlation identifier.
        n_spin (ndarray): Real-space electronic densities per spin channel.
        Nspin (int): Number of spin states.

    Keyword Args:
        dens_threshold (float): Do not treat densities smaller than the threshold.
        exc_only (bool): Only calculate the exchange-correlation energy density.

    Returns:
        tuple[ndarray, ndarray]: Exchange-correlation energy density and potential.
    '''
    if isinstance(xc, str):
        xc = parse_functionals(xc)
    f_exch, f_corr = xc

    # Only use non-zero values of the density
    n = np.sum(n_spin, axis=0)
    nz_mask = np.where(n > dens_threshold)
    n_nz = n[nz_mask]
    # Zeta is only needed for non-zero values of the density
    zeta_nz = get_zeta(n_spin[:, nz_mask])

    # Only import libxc interface if necessary
    if 'libxc' in str(xc):
        from ..extras.libxc import libxc_functional

    # Handle exchange part
    if 'libxc' in f_exch:
        f_exch = f_exch.split(':')[1]
        ex, vx = libxc_functional(f_exch, n_spin, Nspin)
    else:
        if Nspin == 2 and f_exch != 'mock_xc':
            f_exch += '_spin'
        ex_nz, vx_nz = IMPLEMENTED[f_exch](n_nz, zeta=zeta_nz, Nspin=Nspin, exc_only=exc_only)
        # Map the non-zero values back to the right dimension
        ex = np.zeros_like(n)
        ex[nz_mask] = ex_nz
        vx = np.zeros_like(n_spin)
        if not exc_only:
            for spin in range(Nspin):
                vx[spin, nz_mask] = vx_nz[spin]

    # Handle correlation part
    if 'libxc' in f_corr:
        f_corr = f_corr.split(':')[1]
        ec, vc = libxc_functional(f_corr, n_spin, Nspin)
    else:
        if Nspin == 2 and f_corr != 'mock_xc':
            f_corr += '_spin'
        ec_nz, vc_nz = IMPLEMENTED[f_corr](n_nz, zeta=zeta_nz, Nspin=Nspin, exc_only=exc_only)
        # Map the non-zero values back to the right dimension
        ec = np.zeros_like(n)
        ec[nz_mask] = ec_nz
        vc = np.zeros_like(n_spin)
        if not exc_only:
            for spin in range(Nspin):
                vc[spin, nz_mask] = vc_nz[spin]

    return ex + ec, vx + vc


def get_exc(*args, **kwargs):
    '''Get the exchange-correlation energy density.

    This is a convenience function to interface :func:`~eminus.xc.utils.get_xc`.
    '''
    exc, _ = get_xc(*args, **kwargs, exc_only=True)
    return exc


def get_vxc(*args, **kwargs):
    '''Get the exchange-correlation potential.

    This is a convenience function to interface :func:`~eminus.xc.utils.get_xc`.
    '''
    _, vxc = get_xc(*args, **kwargs)
    return vxc


def parse_functionals(xc):
    '''Parse exchange-correlation functional strings to the internal format.

    Args:
        xc (str): Exchange and correlation identifier, separated by a comma.

    Returns:
        list: Exchange and correlation string.
    '''
    # Check for combined aliases
    try:
        xc = ALIAS[xc]
    except KeyError:
        pass

    # Parse functionals
    functionals = []
    for f in xc.split(','):
        if 'libxc' in f or f in IMPLEMENTED.keys():
            f_xc = f
        elif f == '':
            f_xc = 'mock_xc'
        else:
            try:
                f_xc = XC_MAP[f]
            except KeyError:
                log.exception(f'No functional found for "{f}"')
                raise
        functionals.append(f_xc)

    # If only one or no functional has been parsed append with mock functionals
    for i in range(2 - len(functionals)):
        functionals.append('mock_xc')
    return functionals


def get_zeta(n_spin):
    '''Calculate the relative spin polarization.

    Args:
        n_spin (ndarray): Real-space electronic densities per spin channel.

    Returns:
        ndarray: Relative spin polarization.
    '''
    # If only one spin is given return an array of ones as if the density only is in one channel
    if len(n_spin) == 1:
        return np.ones_like(n_spin[0])
    return (n_spin[0] - n_spin[1]) / (n_spin[0] + n_spin[1])


def mock_xc(n, Nspin=1, **kwargs):
    '''Mock exchange-correlation functional with no effect (spin-paired).

    Args:
        n (ndarray): Real-space electronic density.
        Nspin (int): Number of spin states.

    Returns:
        tuple[ndarray, ndarray]: Mock exchange-correlation energy density and potential.
    '''
    zeros = np.zeros_like(n)
    return zeros, np.array([zeros] * Nspin)


IMPLEMENTED = {
    'mock_xc': mock_xc,
    'lda_x': lda_x,
    'lda_x_spin': lda_x_spin,
    'lda_c_pw': lda_c_pw,
    'lda_c_pw_spin': lda_c_pw_spin,
    'lda_c_pw_mod': lda_c_pw_mod,
    'lda_c_pw_mod_spin': lda_c_pw_mod_spin,
    'lda_c_vwn': lda_c_vwn,
    'lda_c_vwn_spin': lda_c_vwn_spin,
    'lda_c_chachiyo': lda_c_chachiyo,
    'lda_c_chachiyo_spin': lda_c_chachiyo_spin,
    'lda_c_chachiyo_mod': lda_c_chachiyo_mod,
    'lda_c_chachiyo_mod_spin': lda_c_chachiyo_mod_spin
}

XC_MAP = {
    # lda_x
    '1': 'lda_x',
    's': 'lda_x',
    'lda': 'lda_x',
    'slater': 'lda_x',
    # lda_c_vwn
    '7': 'lda_c_vwn',
    'vwn': 'lda_c_vwn',
    'vwn5': 'lda_c_vwn',
    # lda_c_pw
    '12': 'lda_c_pw',
    'pw': 'lda_c_pw',
    'pw92': 'lda_c_pw',
    # lda_c_pw_mod
    '13': 'lda_c_pw_mod',
    'pw_mod': 'lda_c_pw_mod',
    'pw92_mod': 'lda_c_pw_mod',
    # lda_c_chachiyo
    '287': 'lda_c_chachiyo',
    'chachiyo': 'lda_c_chachiyo',
    # lda_c_chachiyo_mod
    '307': 'lda_c_chachiyo_mod',
    'chachiyo_mod': 'lda_c_chachiyo_mod'
}

ALIAS = {
    'spw92': 'slater,pw_mod',
    'svwn': 'slater,vwn'
}



#!/usr/bin/env python3
'''GTH file handling.'''
import inspect
import pathlib

import numpy as np

from ..logger import log


def read_gth(atom, charge=None, psp_path=None):
    '''Read GTH files for a given atom.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        atom (str): Atom name.

    Keyword Args:
        charge (int): Valence charge.
        psp_path (str): Path to GTH pseudopotential files. Defaults to installation_path/pade/.

    Returns:
        dict: GTH parameters.
    '''
    if psp_path is None:
        file_path = pathlib.Path(inspect.getfile(inspect.currentframe())).parent
        psp_path = file_path.parent.joinpath('pade')

    if charge is not None:
        f_psp = psp_path.joinpath(f'{atom}-q{charge}')
    else:
        files = sorted(psp_path.glob(f'{atom}-q*'))
        try:
            f_psp = pathlib.Path(files[0])
        except IndexError:
            log.warning(f'There is no GTH pseudopotential in {psp_path} for "{atom}"')
            return mock_gth()
        if len(files) > 1:
            log.info(f'Multiple pseudopotentials found for "{atom}". '
                     f'Continue with "{f_psp.name}".')

    psp = {}
    cloc = np.zeros(4)
    rp = np.zeros(4)
    Nproj_l = np.zeros(4, dtype=int)
    h = np.zeros((4, 3, 3))
    try:
        with open(f_psp, 'r') as fh:
            atom = fh.readline()
            # Skip the first line, since we don't need the atom symbol here. If needed, use
            # psp['atom'] = atom.split()[0]  # Atom symbol
            N_all = fh.readline().split()
            psp['Zion'] = sum([int(N) for N in N_all])  # Ionic charge
            loc = fh.readline().split()
            psp['rloc'] = float(loc[0])  # Range of local Gaussian charge distribution
            # Skip the number of local coefficients, since we don't need it. If needed, use
            # psp['n_c_local'] = int(loc[1])  # Number of local coefficients
            for i, val in enumerate(loc[2:]):
                cloc[i] = float(val)
            psp['cloc'] = cloc  # Coefficients for the local part
            lmax = int(fh.readline().split()[0])
            psp['lmax'] = lmax  # Maximal angular momentum in the non-local part
            for i in range(lmax):
                proj = fh.readline().split()
                rp[i], Nproj_l[i] = float(proj[0]), int(proj[1])
                for k, val in enumerate(proj[2:]):
                    h[i, 0, k] = float(val)
                for j in range(1, Nproj_l[i]):
                    proj = fh.readline().split()
                    for k, val in enumerate(proj):
                        h[i, j, j + k] = float(val)
                # Copy upper triangle elements to the lower triangle
                for jtmp in range(3):
                    for ktmp in range(i, 3):
                        h[i, ktmp, jtmp] = h[i, jtmp, ktmp]
            psp['rp'] = rp  # Projector radius for each angular momentum
            psp['Nproj_l'] = Nproj_l  # Number of non-local projectors
            psp['h'] = h  # Projector coupling coefficients per AM channel
    except FileNotFoundError:
        log.warning(f'There is no GTH pseudopotential for "{f_psp.name}"')
        return mock_gth()
    return psp


def mock_gth():
    '''Create a mock dictionary with all-zeros, for atom species with no pseudopotential file.

    Returns:
        dict: GTH parameters (all zero).
    '''
    psp = {}
    psp['Zion'] = 0
    psp['rloc'] = 0
    psp['cloc'] = np.zeros(4)
    psp['lmax'] = 0
    psp['rp'] = np.zeros(4)
    psp['Nproj_l'] = np.zeros(4, dtype=int)
    psp['h'] = np.zeros((4, 3, 3))
    return psp
